#!/usr/local/bin/perl
#3_read_many_seqs.pl
#Author: Paul Stothard
#
#This script reads multiple DNA sequences (in FASTA format) from a
#file and then prints the name, length, and %GC content of each
#sequence.

use warnings;
use strict;

#Declare an array for use later on in the script. An array is
#essentially a list of text entries or numbers. Perl has many built in
#functions for manipulating arrays. Unlike variables that hold scalar
#values (like text and numbers) array names start with the "@" symbol,
#which tells Perl that the variable is an array. The array we define
#will be empty for now.
my @array_of_names = ();

#Declare another array, which we will use to hold the sequence lengths.
my @array_of_lengths = ();

#Declare an array to hold the %GC content of the sequences.
my @array_of_GC = ();

#Prompt the user for the name of the file to read.
print "Enter the name of the sequence file"
    . " (try short_seqs.txt) and then press enter:\n";
my $file_to_read = <STDIN>;

#Remove the \n from the end of the filename.
chomp($file_to_read);

#Assign a filehandle to the file.
open( my $DNAFILE, $file_to_read )
    or die("Cannot open file '$file_to_read': $!");

#By default, the "<>" operator reads one line at a time. This behavior
#can be changed by changing the value of a built in Perl variable
#"$/". Because we are reading FASTA files, we will assign $/ to
#equal ">". The 'local' keyword should be used when changing the value
#of $/.
local $/ = ">";

#Now read the file in segments, with each segment being the text up to
#and including the next ">" character in the file. Use a "while" loop
#to keep executing the code between the curly braces until
#$sequence_entry can no longer be assigned a value, because the entire
#file has been read.
while ( my $sequence_entry = <$DNAFILE> ) {

    #The first ">" will be encountered before the first sequence title. We
    #want to skip this first segment as it will only contain ">". We will
    #skip any entry that does not contain any text. The "eq" is used for
    #text. It evaluates to true if the two items are the same.

    if ( $sequence_entry eq ">" ) {

        #Go through the next iteration of the while loop using "next".
        next;
    }

    #The entries that contain more than just ">" should hopefully have a
    #title followed by a carriage return. Grab this title using a "regular
    #expression", which is a pattern of characters used to search through
    #text. The following statement uses the matching operator "m//" to
    #find all the text before the first carriage return in the text stored
    #in $sequence_entry. Between the two slashes is the regular expression,
    #which in this case is "^([^\n]+)". There are several characters that
    #can be used in regular expressions to describe the pattern you want
    #to match. In this particular case, the regular expression matches all
    #of the text from the start of the string to, but not including, the
    #first carriage return. The first "^" means "start of string", the
    #[^\n]+ means one or more (the "+" means one or more) characters that
    #are not (the "^" after the opening square bracket means "not")
    #carriage returns (the "\n" means carriage return). The brackets "()"
    #inside of the slashes tell Perl to automatically store the actual
    #text that matched in a special variable that Perl creates
    #automatically. The variable is called "$1". The first line creates an
    #empty variable called $sequence_title. The "if" statement causes the
    #code between the curly braces to be executed, should the contents of
    #the if statement evaluate to be true. The matching operator returns
    #"true" if it finds the pattern described by the regular expression
    #inside the variable it is assigned using "=~". In this case the
    #variable is $sequence_entry. Thus if the text stored in $sequence_entry
    #contains a stretch of text and then a carriage return, $sequence_title
    #is assigned the value stored in $1.
    my $sequence_title = "";
    if ( $sequence_entry =~ m/^([^\n]+)/ ) {
        $sequence_title = $1;
    }
    else {
        $sequence_title = "No title was found!";
    }

    #Now we want to extract the DNA sequence of the sequence record. To do
    #this we can use the substitution operator "s///" to replace all the
    #text before the first carriage return. Between the first two slashes
    #is the regular expression, and between the last two slashes is the
    #text we want to replace the removed text with. By not placing
    #anything between the second two slashes we are essentially removing
    #the text that matches the regular expression (replacing it with
    #nothing). The substitution operator modifies the variable it is
    #applied to using the binding operator.
    $sequence_entry =~ s/^[^\n]+//;

    #Now we will add these values to the arrays we declared near the top
    #of this script. The "push" function receives two arguments: an array
    #and an element to add to the array. In this case we add the value of
    #$sequence_title to the @array_of_names array.
    push( @array_of_names, $sequence_title );

    #Add the length of the sequence to the array. The first line gets rid
    #of non-DNA characters using the substitution operator, and the second
    #uses the "length" function to determine the length of the text. This
    #length value is then pushed onto the end of the array. The regular
    #expression in the first line matches any characters other than those
    #between A and Z. The "g" after the second slash is called the "global
    #modifier". It causes the substitution operator to replace every
    #occurrence of the pattern, instead of just the first. The "i"
    #is another modifier, and it is used to indicate that the text in the
    #regular expression represents uppercase and lowercase text. The
    #substitution expression below could be rewritten as s/[^A-Za-z]//g.
    $sequence_entry =~ s/[^A-Z]//gi;
    push( @array_of_lengths, length($sequence_entry) );

    #Determine the percent GC content of the sequence. First we will store
    #the length of the sequence in a variable called $sequence_length.
    my $sequence_length = length($sequence_entry);

    #Now we will use the substitution operator in a slightly different
    #way. Whenever a substitution is performed, the contents of the
    #variable bound to the substitution operator (with the "=~") are
    #modified. The substitution operator also returns the number of
    #replacements it made. You may not always be interested in this value,
    #but if you are you can simply store it in a variable. In this case we
    #will use a variable called $number_of_replacements.
    my $number_of_replacements = $sequence_entry =~ s/[GCgc]//g;

    #To determine %GC content, we divide the number of replacements made,
    #which should be the same as the number of G's or C's, by the length
    #of the sequence and multiply the result by 100. Although the brackets
    #are not necessary in this case, they improve readability.
    my $GC_content = ( $number_of_replacements / $sequence_length ) * 100;

    #Use the "sprintf" function of perl to round the value to two decimal
    #places. The "%.2f" is called the format control string. It tells the
    #sprintf function to format the value it receives (in this case the
    #value in $GC_content) as a floating point number rounded to two
    #decimal places. The formatted text is then placed back into
    #$GC_content, because of the "$GC_content =".
    $GC_content = sprintf( "%.2f", $GC_content );

    #Now push this value onto the end of the array for GC content.
    push( @array_of_GC, $GC_content );

    #All the statements between the "while" and the following curly brace
    #get executed for each sequence entry. This block of code contains
    #several variables that are first declared inside this block
    #($sequence_title for example). These variables can only be used inside
    #the block, and they are assigned new values each time the block is
    #executed. The arrays we use were declared outside of the while loop
    #block, at the start of the program. This allows the arrays to be used
    #anywhere in the script. Each time this block of code executes, new
    #values are added to the end of the three arrays, such that the length
    #of each array increases by 1.

}

#Once you are done with a file, close the filehandle.
close($DNAFILE) or die("Cannot close file : $!");

#Now we want to display the contents of the arrays. We will use a
#"for" loop to cycle through the array elements. The loop increases
#the value of the variable $i each iteration. The value starts at
#zero, which is the index used to access the first element in an
#array. The loop continues until $i is equal to the length of the
#array minus 1, which is always the index of the last element in an
#array. When accessing a value in an array, you use the name of the
#array, but you replace the "@" with a "$", followed by the position
#number of the element in square brackets. For example, to refer to
#the first sequence name in the array @array_of_names, you would use
#$array_of_names[0].
for ( my $i = 0; $i < scalar(@array_of_names); $i = $i + 1 ) {
    print "Name: $array_of_names[$i]\n";
    print "Length: $array_of_lengths[$i] bp\n";
    print "\%GC: $array_of_GC[$i]\n";
    print "-----------------\n";
}

#Use the built in "sort" function. Sort does not sort the array it is
#given--it returns a copy of the array in sorted form, which can be
#stored in a new array variable.
my @sorted_array = sort { $a <=> $b } @array_of_GC;

#Print the smallest %GC value, which will be the first value in the
#sorted array.
print "The smallest \%GC value is: $sorted_array[0]\n";

#To print the largest %GC value, reverse the array and print the first
#element.
my @reversed_sorted_array = reverse(@sorted_array);
print "The largest \%GC value is: $reversed_sorted_array[0]\n";

#Once your script is running correctly using short_seqs.txt, try using
#sars.txt as the input to see how your script performs. Try modifying
#this program so that the name of the sequence (or sequences) with the
#largest %GC value is also printed.

#To run this script, use the following command:
#perl 3_read_many_seqs.pl
