#!/usr/local/bin/perl
#4_translate.pl
#Author: Paul Stothard
#
#This script reads multiple DNA sequences (in FASTA format) from a
#file and then translates the DNA into protein in reading frame 1.

use warnings;
use strict;

#Declare an array for use later on in the script, which we will use to
#hold sequence names..
my @array_of_names = ();

#Declare an array to hold the protein translation of each sequence.
my @array_of_translations = ();

#Declare a hash table for associating codons with amino acids. A hash
#table is sometimes called an associative array. It is an array in
#which words are assigned to a particular value. The word (also called
#the key), can be used at a later time to obtain the value. Here we
#will create a hash table that we will use for translating DNA into
#protein. Every possible codon will serve as a key, and the
#corresponding amino acid will be the value that is associated with
#the key. The name of a hash table in perl always begins with a
#"%". To make the contents we are placing in the hash easy to read,
#the statement has been spread over many lines. The Perl interpreter
#ignores blank spaces (it simply looks for the semicolon that marks
#the end of the statement). Because we are just specifying the
#lowercase codons, we will have to remember to convert the DNA
#sequences to lowercase (using a Perl function).
my %translation_hash = 
    (gca => "A", gcg => "A", gct => "A", gcc => "A", gcn => "A",
     tgc => "C", tgt => "C",
     gat => "D", gac => "D",
     gaa => "E", gag => "E",
     ttt => "F", ttc => "F",
     gga => "G", ggg => "G", ggc => "G", ggt => "G", ggn => "G",
     cat => "H", cac => "H",
     ata => "I", att => "I", atc => "I",
     aaa => "K", aag => "K",
     cta => "L", ctg => "L", ctt => "L", ctc => "L", ctn => "L", tta => "L", ttg => "L",
     atg => "M",
     aat => "N", aac => "N",
     cca => "P", cct => "P", ccg => "P", ccc => "P", ccn => "P",
     caa => "Q", cag => "Q",
     cga => "R", cgg => "R", cgc => "R", cgt => "R", cgn => "R",
     aga => "R", agg => "R",
     tca => "S", tcg => "S", tcc => "S", tct => "S", tcn => "S",
     agc => "S", agt => "S",
     aca => "T", acg => "T", acc => "T", act => "T", acn => "T",
     gta => "V", gtg => "V", gtc => "V", gtt => "V", gtn => "V",
     tgg => "W",
     tat => "Y", tac => "Y",
     tag => "*", taa => "*", tga => "*");

#Prompt the user for the name of the file to read.
print "Enter the name of the sequence file"
    . " (try short_seqs.txt) and then press Enter:\n";
my $file_to_read = <STDIN>;

#Remove the \n from the end of the filename.
chomp($file_to_read);

#Assign a filehandle to the file.
open( my $DNAFILE, $file_to_read )
    or die("Cannot open file '$file_to_read': $!");

#Because we are reading FASTA files, we will assign "$/" to equal ">".
local $/ = ">";

#Now read the file in segments, with each segment being the text up to
#and including the next ">" character in the file. Use a "while" loop
#to keep executing the code between the curly braces until
#$sequence_entry can no longer be assigned a value, because the entire
#file has been read.
while ( my $sequence_entry = <$DNAFILE> ) {

    #The first ">" will be encountered before the first sequence title. We
    #want to skip this first segment as it will only contain ">". We will
    #skip any entry that does not contain any text. The "eq" is used for
    #text. It evaluates to true if the two items are the same.
    if ( $sequence_entry eq ">" ) {

        #Go through the next iteration of the while loop using "next".
        next;
    }

    #The entries that contain more than just ">" should hopefully have a
    #title followed by a carriage return. Grab this title using a "regular
    #expression", which is a pattern of characters used to search through
    #text. The following statement uses the matching operator "m//" to
    #find all the text before the first carriage return in the text stored
    #in $sequence_entry. Between the two slashes is the regular expression,
    #which in this case is "^([^\n]+)". There are several characters that
    #can be used in regular expressions to describe the pattern you want
    #to match. In this particular case, the regular expression matches all
    #of the text from the start of the string to, but not including, the
    #first carriage return. The first "^" means "start of string", the
    #[^\n]+ means one or more (the "+" means one or more) characters that
    #are not (the "^" after the opening square bracket means "not")
    #carriage returns (the "\n" means carriage return). The brackets "()"
    #inside of the slashes tell Perl to automatically store the actual
    #text that matched in a special variable that Perl creates
    #automatically. The variable is called "$1". The first line creates an
    #empty variable called $sequence_title. The "if" statement causes the
    #code between the curly braces to be executed, should the contents of
    #the if statement evaluate to be true. The matching operator returns
    #"true" if it finds the pattern described by the regular expression
    #inside the variable it is assigned using "=~". In this case the
    #variable is $sequence_entry. Thus if the text stored in $sequence_entry
    #contains a stretch of text and then a carriage return, $sequence_title
    #is assigned the value stored in $1.
    my $sequence_title = "";
    if ( $sequence_entry =~ m/^([^\n]+)/ ) {
        $sequence_title = $1;
    }
    else {
        $sequence_title = "No title was found!";
    }

    #Now we want to extract the DNA sequence of the sequence record. To do
    #this we can use  the substitution operator "s///" to replace all the
    #text before the first carriage return. Between the first two slashes
    #is the regular expression, and between the last two slashes is the
    #text we want to replace the removed text with. By not placing
    #anything between the second two slashes we are essentially removing
    #the text that matches the regular expression (replacing it with
    #nothing). The substitution operator modifies the variable it is
    #applied to using the binding operator.
    $sequence_entry =~ s/^[^\n]+//;

    #Now we will add these values to the arrays we declared near the top
    #of this script. The "push" function receives two arguments: an array
    #and an element to add to the array. In this case we add the value of
    #$sequence_title to the @array_of_names array.
    push( @array_of_names, $sequence_title );

    #Prepare the DNA for our translation hash table. The regular
    #expression in the first line matches any characters other than "G",
    #"A", "T", "C", "N", "g", "a", "t", "c", and "n". The "g" after the
    #second slash is called the "global modifier". It causes the
    #substitution operator to replace every occurrence of the pattern,
    #instead of just the first.
    $sequence_entry =~ s/[^GATCNgatcn]//g;

    #Now convert the sequence to lowercase text using the function
    #"lc". This function doesn't change the original string--it returns
    #the contents of the variable it receives, but in lowercase form. By
    #assigning the returned value to $sequence_entry (using "="), we
    #actually are replacing the contents of $sequence_entry with the new
    #lowercase form of the text.
    $sequence_entry = lc($sequence_entry);

    #Now we will declare an array that we can use to hold the amino acids
    #of the protein that is produced as the DNA sequence is translated.
    my @growing_protein = ();

    #Now we need to read the DNA sequence in chunks of three bases. A for
    #loop with an increment of 3 can be used. The loop continues while $i
    #is less than or equal to the length of the DNA sequence - 3. This
    #condition ensures that the "substr" function (described below) will
    #be able to return a complete codon.
    for ( my $i = 0; $i <= ( length($sequence_entry) - 3 ); $i = $i + 3 ) {

      #Use the "substr" function to obtain the codon we want to
      #translate. It receives three arguments: the string to work on, the
      #position of the first character to remove, and the number of
      #characters to return. We want substr to return characters from
      #$sequence_entry, starting at the position stored in $i. We want three
      #characters to be returned. These will be stored in the variable $codon;
        my $codon = substr( $sequence_entry, $i, 3 );

        #Now check if the hash table contains the codon, and if it does, get
        #the amino acid from the hash table. The "exists" function returns
        #true if the hash table contains the given key value, which in this
        #case is the value stored in $codon. When dealing with specific
        #entries in the hash table, the name of the table starts with "$"
        #instead of "%", and the key that is being used to get a value is
        #placed between curly braces.
        if ( exists( $translation_hash{$codon} ) ) {

        #The codon is in the hash table, so the amino acid that corresponds to
        #this codon can be pushed onto the end of the @growing_protein array.
            push( @growing_protein, $translation_hash{$codon} );
        }
        else {

           #The code in this else block is executed if the hash table doesn't
           #contain the codon as a key. This could be the case for codons that
           #contain several "n" bases. We will translate these to "X" to mean
           #unknown.

            push( @growing_protein, "X" );
        }

        #The code between the "for" statement and the curly brace after this
        #comment gets executed once for each codon in the sequence.

    }    #------------END OF "FOR" LOOP

    #Now the for loop has been used to make an array of amino acids. We
    #can easily combine these into a string using the "join"
    #function. This function is given a separator string (in this case we
    #do not want the items separated by anything, so we give the function
    #an empty string ""), followed by an array. The function joins the
    #elements and returns the resulting text. In this case we store the
    #text in a new variable called $joined_amino_acids.
    my $joined_amino_acids = join( "", @growing_protein );

    #The text contained in $joined_amino_acids can now be added on to our
    #array of protein translations.
    push( @array_of_translations, $joined_amino_acids );

    #All the statements between the "while" and the following curly brace
    #get executed for each sequence entry. This block of code contains
    #several variables that are first declared inside this block
    #($sequence_title for example). These variables can only be used inside
    #the block, and they are assigned new values each time the block is
    #executed. The arrays we use were declared outside of the while loop
    #block, at the start of the program. This allows the arrays to be used
    #anywhere in the script. Each time this block of code executes, new
    #values are added to the end of the three arrays, such that the length
    #of each array increases by 1.

}    #---------------END OF "WHILE" LOOP

#Once you are done with a file, close the filehandle.
close($DNAFILE) or die("Cannot close file : $!");

#Now we want to display the contents of the arrays. We will use a
#"for" loop to cycle through the array elements. The loop increases
#the value of the variable $i each iteration. The value starts at
#zero, which is the index used to access the first element in an
#array. The loop continues until $i is equal to the length of the
#array minus 1, which is always the index of the last element in an
#array. When accessing a value in an array, you use the name of the
#array, but you replace the "@" with a "$", followed by the position
#number of the element in square brackets. For example, to refer to
#the first sequence name in the array @array_of_names, you would use
#$array_of_names[0].
for ( my $i = 0; $i < scalar(@array_of_names); $i = $i + 1 ) {
    print "Name: $array_of_names[$i]\n";
    print "Translation: $array_of_translations[$i]\n";
    print "-----------------\n";
}    #---------------END OF "FOR" LOOP

#Once your script is working on the short_seqs.txt, test it on sars.txt
#to see how it performs.

#To run this script, use the following command:
#perl 4_translate.pl
