#!/usr/local/bin/perl
#5_rev_comp.pl
#Author: Paul Stothard
#
#This script reads a single DNA sequence (in FASTA format) from a file
#and then converts the DNA sequence into its reverse-complement form.

use warnings;
use strict;

#Prompt the user for the name of the file to read.
print "Enter the name of the sequence file"
    . " (try sars.txt) and then press enter:\n";
my $file_to_read = <STDIN>;

#Remove the \n from the end of the filename.
chomp($file_to_read);

#Assign a filehandle to the file.
open( my $DNAFILE, $file_to_read )
    or die("Cannot open file '$file_to_read': $!");

#If we set the built in variable "$/" to "undef", the diamond operator
#will read the entire file. We will store the file contents in the
#variable $sequence_entry.
local $/ = undef;
my $sequence_entry = <$DNAFILE>;

#Now use the matching operator to obtain the title of the sequence. A
#">" will be at the start of the title line, so we can include the ">"
#in our regular expression. If a title is not found, the statements in
#the "else" block are executed. The "die" statement stops the program
#and prints the message advising the user of the problem.
my $sequence_title = "";
if ( $sequence_entry =~ m/(>[^\n]+)/ ) {
    $sequence_title = $1;

    #Remove the ">" from the sequence title using the substitution
    #operator.
    $sequence_title =~ s/>//;
}
else {
    die("A FASTA sequence title was not found.");
}

#Once you are done with a file, close the filehandle.
close($DNAFILE) or die("Cannot close file : $!");

#The contents of $sequence_entry are not changed by the previous block
#of code. To obtain the DNA sequence portion of $sequence_entry, we
#can remove the title line using the substitution operator. The
#substitution operator modifies the variable it is applied to using
#the binding operator "=~".
$sequence_entry =~ s/>[^\n]+//;

#Now convert the sequence to lowercase text. We will use a different
#method this time--the translation operator "tr". It receives two sets
#of characters. The characters between the first two slashes are
#called the "search characters". These characters are found in the
#contents of the variable bound to the tr operator. The characters
#between the second and third slash are called the replacement
#characters. The "tr" operator scans the text and replaces every
#instance of the search characters with their corresponding characters
#in the replacement characters. In the line below, only the DNA
#characters are changed to lowercase.
$sequence_entry =~ tr/GATCN/gatcn/;

#Remove any blank spaces, numbers, and non-DNA characters from the
#remaining text. The regular expression (between the slashes) matches
#any characters other than "g", "a", "t", "c", and "n". The "g" after
#the second slash is called the "global modifier". It causes the
#substitution operator to replace every occurrence of the pattern,
#instead of just the first.
$sequence_entry =~ s/[^gatcn]//g;

#Store a copy of the DNA sequence for use later in the program, as we
#will be modifying the version stored in $sequence_entry.
my $original_sequence = $sequence_entry;

#To convert the DNA sequence to its complement, we can use the
#translation operator again.
$sequence_entry =~ tr/gatcn/ctagn/;

#Now use the reverse function to reverse the sequence.
my $reverse_comp_sequence = reverse($sequence_entry);

#Now print some information about the original sequence and the
#reverse-complement. The "." that appears between the items in the
#"print" statement joins the values into one string of text. The
#results of several functions are included in the print statement
#using "." in the below code.
print "The reverse-complement results for " . $sequence_title . " are:\n";
print "Original length: " . length($original_sequence) . " bp.\n";
print "Length after reverse-complement: "
    . length($reverse_comp_sequence)
    . " bp.\n";
print "\n";
print "The original sequence begins with: "
    . substr( $original_sequence, 0, 5 ) . "\n";

#Note that "substr" can accept a negative value for the starting
#position. Below we use "-5", which causes substr to return the
#characters starting at 5 characters from the end of the text.
print "and ends with: " . substr( $original_sequence, -5, 5 ) . ".\n";
print "The reverse-complement begins with: "
    . substr( $reverse_comp_sequence, 0, 5 ) . "\n";
print "and ends with: " . substr( $reverse_comp_sequence, -5, 5 ) . ".\n";
print "\n";

#Now we will use the translation operator to count the number of each
#base type in the sequences. When a translation is assigned to a
#variable, the number of replacements made is stored. The brackets
#clarify that Perl should perform the translation operation first.
my $g = ( $original_sequence =~ tr/g/G/ );
my $a = ( $original_sequence =~ tr/a/A/ );
my $t = ( $original_sequence =~ tr/t/T/ );
my $c = ( $original_sequence =~ tr/c/C/ );
print "The original sequence contains $g g's, $a a's, $t t's, and $c c's.\n";

#reuse the same four variables for the reverse-complement.
$g = ( $reverse_comp_sequence =~ tr/g/G/ );
$a = ( $reverse_comp_sequence =~ tr/a/A/ );
$t = ( $reverse_comp_sequence =~ tr/t/T/ );
$c = ( $reverse_comp_sequence =~ tr/c/C/ );
print "The reverse-complement contains $g g's, $a a's, $t t's, and $c c's.\n";

#To run this script, use the following command:
#perl 5_rev_comp.pl
