#!/usr/local/bin/perl
#6_orfs.pl
#Author: Paul Stothard
#
#This script reads a single DNA sequence (in FASTA format) from a file
#and then finds the open reading frames (ORFs) on the direct strand.

use warnings;
use strict;

#Prompt the user for the name of the file to read.
print "Enter the name of the sequence file"
    . " (try medium_seq.txt) and then press enter:\n";
my $file_to_read = <STDIN>;

#Remove the \n from the end of the filename.
chomp($file_to_read);

#Assign a filehandle to the file.
open( my $DNAFILE, $file_to_read )
    or die("Cannot open file '$file_to_read': $!");

#If we set the built in variable "$/" to "undef", the diamond operator
#will read the entire file. We will store the file contents in the
#variable $sequence_entry.
local $/ = undef;
my $sequence_entry = <$DNAFILE>;

#Now use the matching operator to obtain the title of the sequence. A
#">" will be at the start of the title line, so we can include the ">"
#in our regular expression. If a title is not found, the statements in
#the "else" block are executed. The "die" statement stops the program
#and prints the message advising the user of the problem.
my $sequence_title = "";
if ( $sequence_entry =~ m/(>[^\n]+)/ ) {
    $sequence_title = $1;

    #Remove the ">" from the sequence title using the substitution operator.
    $sequence_title =~ s/>//;
}
else {
    die("A FASTA sequence title was not found.");
}

#Once you are done with a file, close the filehandle.
close($DNAFILE) or die("Cannot close file : $!");

#The contents of $sequence_entry are not changed by the previous block
#of code. To obtain the DNA sequence portion of $sequence_entry, we can
#remove the title line using the substitution operator. The
#substitution operator modifies the variable it is applied to using
#the binding operator "=~".
$sequence_entry =~ s/>[^\n]+//;

#Now convert the sequence to lowercase text. We will use a different
#method this time--the translation operator "tr". It receives two sets
#of characters. The characters between the first two slashes are
#called the "search characters". These characters are found in the
#contents of the variable bound to the tr operator. The characters
#between the second and third slash are called the replacement
#characters. The "tr" operator scans the text and replaces every
#instance of the search characters with their corresponding characters
#in the replacement characters. In the line below, only the DNA
#characters are changed to lowercase.
$sequence_entry =~ tr/GATCN/gatcn/;

#Remove any blank spaces, numbers, and non-DNA characters from the
#remaining text. The regular expression (between the slashes) matches
#any characters other than "g", "a", "t", "c", and "n". The "g" after
#the second slash is called the "global modifier". It causes the
#substitution operator to replace every occurrence of the pattern,
#instead of just the first.
$sequence_entry =~ s/[^gatcn]//g;

#Declare an array to hold the ranges of ORFs found in the sequence.
my @array_of_ORFs = ();

#Declare arrays to hold the positions of start and stop codons for
#each reading frame.
my @starts_RF1 = ();
my @starts_RF2 = ();
my @starts_RF3 = ();
my @stops_RF1  = ();
my @stops_RF2  = ();
my @stops_RF3  = ();

#Find all the start codons in the sequence using the matching
#operator. The regular expression is "atg". The "g" after the second
#slash is the global modifier, and it causes Perl to match every
#instance of the regular expression. The "i" tells perl to ignore the
#case of the text when matching. Each time an atg is found in the
#sequence, the statements inside the while loop are executed. For
#example, if a sequence contains 10 atg's, the statements inside the
#while loop will execute 10 times. The first line inside the while
#loop uses the built in Perl function "pos". It returns the offset of
#where the last global regular expression search left off, in this
#case for the variable $sequence_entry. For example, if the sequence
#being searched is "atgatg", the "pos" function will return "3" for
#the first iteration, "6" for the second iteration. We subtract 3 from
#the pos value to convert it to the position of the "a" in the atg.
while ( $sequence_entry =~ m/atg/gi ) {
    my $match_position = pos($sequence_entry) - 3;

    #These conditional statements are used to determine the reading frame
    #that the atg is in. The modulus operator "%" returns the remainder
    #after $match_position is divided by 3. If the value is equal to zero
    #the codon is in the first reading frame.
    if ( ( $match_position % 3 ) == 0 ) {
        push( @starts_RF1, $match_position );
    }

    #If it is not in reading frame 1, see if it is in reading frame 2. To
    #do this see if ($match_position + 2) % 3 is equal to zero.
    elsif ( ( ( $match_position + 2 ) % 3 ) == 0 ) {
        push( @starts_RF2, $match_position );
    }

    #If it is not in reading frame 1 or 2, it must be in reading frame 3.
    else {
        push( @starts_RF3, $match_position );
    }
}

#Now use the same strategy to find the positions and reading frames of
#stop codons.
while ( $sequence_entry =~ m/tag|taa|tga/gi ) {
    my $match_position = pos($sequence_entry);
    if ( ( $match_position % 3 ) == 0 ) {
        push( @stops_RF1, $match_position );
    }
    elsif ( ( ( $match_position + 2 ) % 3 ) == 0 ) {
        push( @stops_RF2, $match_position );
    }
    else {
        push( @stops_RF3, $match_position );
    }
}

#We have the positions of start and stop codons for each reading
#frame, now we need to convert these to ORFs. First declare some
#variables that we will need.
my $codon_range    = "";
my $start_position = 0;
my $stop_position  = 0;

#Start by finding ORFs in reading frame 1. The "reverse" function
#returns a copy of the array it receives, but will the elements in
#reverse order. We are reversing the arrays containing the starts and
#stops so that the "pop" function returns the start and stop codons in
#ascending order.
@starts_RF1 = reverse(@starts_RF1);
@stops_RF1  = reverse(@stops_RF1);

#The statements in the while block are repeated until all the start
#positions have been examined.
while ( scalar(@starts_RF1) > 0 ) {
    $codon_range = "";

    #The "pop" function removes the last element from the array it
    #receives and returns the element. In this example we are storing the
    #element in $start_position. The array @starts_RF1 is one element
    #shorter after the pop operation.
    $start_position = pop(@starts_RF1);

    #If the start position is less than $stop_position, skip the remaining
    #statements and pop off the next start position. This skips atg's that
    #are inside of the last ORF. For the first iteration, $stop_position is
    #set to zero. For additional iterations, $stop_position is set to the
    #position of the last ORF-ending stop codon.
    if ( $start_position < $stop_position ) {
        next;
    }

    #Now look for stops. Pop stop positions off of the array of stop
    #positions. Keep popping off stop positions until one is found that is
    #located after the start position. Once one is found, the "last"
    #statement causes immediate exit from the inner while loop. Program
    #execution continues with the first statement after the while loop.
    while ( scalar(@stops_RF1) > 0 ) {
        $stop_position = pop(@stops_RF1);
        if ( $stop_position > $start_position ) {
            last;
        }
    }

    #A stop position may not exist between the start codon and the end of
    #the sequence. If this is the case, the contents of the below if
    #statement are executed.
    if ( $stop_position <= $start_position ) {

       #We want to set $stop_position to the end of the last complete codon in
       #reading frame 1.
        $stop_position
            = length($sequence_entry) - ( length($sequence_entry) % 3 );

        #Construct a description of the ORF to add to @array_of_ORFs
        $codon_range = "+1 " . $start_position . ".." . $stop_position;
        push( @array_of_ORFs, $codon_range );
        last;
    }

    #If a stop position was found before the end of the sequence, the
    #statements inside the following else will execute.
    else {

        #Construct a description of the ORF to add to @array_of_ORFs.
        $codon_range = "+1 " . $start_position . ".." . $stop_position;
        push( @array_of_ORFs, $codon_range );
    }
}

#Reset $stop_position to 0 for first iteration of while loop. Use same
#while loop strategy for reading frame 2 ORFs.
$stop_position = 0;
@starts_RF2    = reverse(@starts_RF2);
@stops_RF2     = reverse(@stops_RF2);
while ( scalar(@starts_RF2) > 0 ) {
    $codon_range    = "";
    $start_position = pop(@starts_RF2);
    if ( $start_position < $stop_position ) {
        next;
    }
    while ( scalar(@stops_RF2) > 0 ) {
        $stop_position = pop(@stops_RF2);
        if ( $stop_position > $start_position ) {
            last;
        }
    }
    if ( $stop_position <= $start_position ) {

       #We want to set $stop_position to the end of the last complete codon in
       #reading frame 2.
        $stop_position = length($sequence_entry)
            - ( ( length($sequence_entry) + 2 ) % 3 );
        $codon_range = "+2 " . $start_position . ".." . $stop_position;
        push( @array_of_ORFs, $codon_range );
        last;
    }
    else {
        $codon_range = "+2 " . $start_position . ".." . $stop_position;
        push( @array_of_ORFs, $codon_range );
    }
}

#Reset $stop_position to 0 for first iteration of while loop. Use same
#while loop strategy for reading frame 3 ORFs.
$stop_position = 0;
@starts_RF3    = reverse(@starts_RF3);
@stops_RF3     = reverse(@stops_RF3);
while ( scalar(@starts_RF3) > 0 ) {
    $codon_range    = "";
    $start_position = pop(@starts_RF3);
    if ( $start_position < $stop_position ) {
        next;
    }
    while ( scalar(@stops_RF3) > 0 ) {
        $stop_position = pop(@stops_RF3);
        if ( $stop_position > $start_position ) {
            last;
        }
    }
    if ( $stop_position <= $start_position ) {

       #We want to set $stop_position to the end of the last complete codon in
       #reading frame 3.
        $stop_position = length($sequence_entry)
            - ( ( length($sequence_entry) + 1 ) % 3 );
        $codon_range = "+3 " . $start_position . ".." . $stop_position;
        push( @array_of_ORFs, $codon_range );
        last;
    }
    else {
        $codon_range = "+3 " . $start_position . ".." . $stop_position;
        push( @array_of_ORFs, $codon_range );
    }
}

#Print the sequence title and length. When a slash is followed by a
#quotation mark, the print statement displays the quotation mark,
#instead of interpreting the quotation mark as the end of the text.
print "Results for \"$sequence_title\"\n";
print "Length = " . length($sequence_entry) . " bp\n";

#The "foreach" loop can be used to execute a block of code once for
#each item in an array. Inside the block of code, you can refer to the
#current array item using the special variable "$_", which Perl
#creates automatically.
foreach (@array_of_ORFs) {

    #Use the matching operator to copy parts of the ORF range into
    #variables. The regular expression used below means the following:
    #match a plus or minus sign followed by a single digit, then a blank
    #space, then one or more digits, then two periods, then one or more
    #digits. The round brackets cause Perl to store the text that actually
    #matches the enclosed parts of the regular expression. Perl stores
    #these in $1, $2, etc.
    $_ =~ m/([\+\-]\d)\s(\d+)\.\.(\d+)/;

    #Print the range of the ORF. 1 is added to the start codon positions
    #when they are displayed because we have been referring to the first
    #base as base 0 instead of base 1.
    print "Reading frame $1 " . ( $2 + 1 ) . " to " . $3;

    #Print the length of the ORF.
    print ", Length: " . ( $3 - $2 ) . ".\n";
}

#Once your script is working using medium_seq.txt, try using sars.txt
#as the input. Also, try creating a single subroutine to replace the
#three duplicated code blocks.

#To run this script, use the following command:
#perl 6_orfs.pl
